#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


import os
from IPy import IP
from cStringIO import StringIO
from net.defs import ip_protocols, PROTO_TCP, PROTO_UDP, PROTO_ICMP
import nunat

IPTABLES="$ipt"

D_ACCEPT='ACCEPT'
D_REJECT='REJECT'
D_DROP='DROP'
D_ULOG='ULOG'
D_LOG='LOG'
D_QUEUE='QUEUE'
D_NFQUEUE='NFQUEUE'
D_CONNMARK='CONNMARK'
D_MARK='MARK'
D_DEFAULT=D_DROP


def gen_rule(*args):
    return " ".join(args)

def gen_rule2(l):
    return gen_rule(*l)

def make_target(s):
    if s=='accept':
        return D_ACCEPT
    if s=='reject':
        return D_REJECT
    if s=='ulog':
        return D_ULOG
    if s=='log':
        return D_LOG
    return D_DROP

def _make_syn(target,h):
    """returns chain that fits well to the target of the iptables command
        example: for accept target: we want to match syn packets, the others are matched by the
        RELATED/ESTABLISHED rules
    """
    s=target
    res=''
    if s == D_ACCEPT:
        res +='-m state --state NEW'
        if h.has_key('proto') and h['proto']=='tcp':
            res+=' --syn'
    return res

class FWipt:
    """Used to generate iptables rules from a firewall object that contains rules
    """
    def __init__(self,thefw,logtype='log', queue=None):
        self.__fw=thefw
        self.__logtype='LOG'
        self.__logprefix='--log-prefix'
        self.__logspace=' ' #one space
        self.queue_type = D_QUEUE
        self.queue_num = None
        if thefw.queue:
            queue = thefw.queue
        if queue:
            self.queue_type = D_NFQUEUE
            self.queue_num = queue
        if logtype=='ulog':
            self.__logtype='ULOG'
            self.__logprefix='--ulog-prefix'
            self.__logspace=''

    def dumb_rules(cls,rules,chain):
        res = ['#ERROR_DISPATCH %s%s'%(chain,os.linesep)]
        sacl = ''
        for r in rules:
            stmp=r.aclname
            if(stmp!=sacl):
                sacl=stmp
                sc='#'+sacl+' (acl '+r.aclid+')'+os.linesep
                res.append(sc)
            s=str(RulePrintIptSimple(r,fw=self.__fw,chain=chain,add=_make_syn(make_target(r.decision),{'proto':r.proto().get('proto')}),target=make_target(r.decision),ulog_prefix=r.ulog_prefix))+os.linesep
            res.append('#'+s)
        return res
    dumb_rules = classmethod(dumb_rules)

    def create_ipt_chains(self):
        """Generate rules to create firewall chains: DMZ-LAN
        """
        l=self.__fw.chains(all=1,inout=1)
        s=''
        s2=''
        acc=[]
        for n in l:
            if self.__fw.chain_input(n) and self.__fw.manage_input is False: continue
            if self.__fw.chain_output(n) and self.__fw.manage_output is False: continue
            #st=IPTABLES+' -N '+n['name']+os.linesep
            st=':'+n['name']+' -'+os.linesep
            if not st in acc: #hack for multiple INTERNET (external zones)
                s+=st
            acc.append(st)
        return s

    def __gen_global_accept(self, rescue, *where):
        state = 'RELATED'
        res = ''
        # Deleted on 2006-01-22
        #if rescue == 1:
        state = 'ESTABLISHED,%s'%state
        for ch in where:
            res = res + gen_rule('-A %s -m state --state %s -j ACCEPT'%(ch,state)) + os.linesep
        return res

    def __gen_default_log_drop(self, *where):
        res = ''
        for ch in where:
            res = res + gen_rule('-A %s -j %s %s "DFT_%s_DROP%s"'%(ch, self.__logtype, self.__logprefix, ch, self.__logspace))+ os.linesep
            res = res + gen_rule('-A %s -j DROP'%(ch))+os.linesep
        return res
    def __gen_default_invalid_drop(self, *where):
        res = ''
        for ch in where:
            res = res + gen_rule('-A %s -m state --state INVALID -j DROP'%(ch)) + os.linesep
        return res

    def gen_loopback_default_rules(self,input=True, output=True):
        s = ''
        if input: s += gen_rule('-A INPUT -i lo -j ACCEPT%s'%os.linesep)
        if output: s += gen_rule('-A OUTPUT -o lo -j ACCEPT%s'%os.linesep)
        return s

    def connect_chains(self,rescue=0):
        """Generate rules that connect firewall streams into chains
           => dispatch rules
        """
        l=self.__fw.chains(all=1,inout=1)
        s=''
        svpn=''
        estrel=''
        sinternet = sinout_internet = ''
        sinput = ''
        soutput = ''
        lwhere = ['FORWARD']
        if self.__fw.manage_input: lwhere.append('INPUT')
        if self.__fw.manage_output: lwhere.append('OUTPUT')
        estrel = self.__gen_global_accept(rescue,*lwhere)
        h_forward = {}
        h_input = {}
        l_output = []
        h_internet = {}
        h_ininternet = {}
        l_outinternet = []
        for n in l:
            #don't generate input/output rules here if are not supposed to use it
            if self.__fw.chain_input(n) and self.__fw.manage_input is False: continue
            if self.__fw.chain_output(n) and self.__fw.manage_output is False: continue
            has_vpn = False
            has_internet = False
            # INPUT/OUTPUT INTEGRATION: change chain name here -> OK
            where = 'FORWARD'
            input = output = False
            if self.__fw.chain_input(n):
                where = 'INPUT'
                input = True
            if self.__fw.chain_output(n):
                where = 'OUTPUT'
                output = True
            inout = False
            if input and output: inout = True
            t=['-A %s'%(where)]
            #nsrc = None
            #ndst = None
##            nsrc=n['net_src']
##            ndst=n['net_dst']
##            IP_nsrc = nsrc.addr()
##            IP_ndst = ndst.addr()
##            # test if net is default route network for firewall
            if n['type'] != "output" and not self.__fw.is_default_net(n['net_src']):
                t.extend(['-s',str(n['net_src'].ip())])
            # TODO: why also output and not just forward ???
            if n['type'] != "input" and not self.__fw.is_default_net(n['net_dst']):
                t.extend(['-d',str(n['net_dst'].ip())])
            # vpn rules
            if (not input and not output) and (n['net_src'].is_vpn() or n['net_dst'].is_vpn()):
                has_vpn=True
                t.extend(['-m policy --pol ipsec --mode tunnel'])
                #remote vpn network is source of the connexion
                # net_dst:
                conn = n['net_dst'].get_conn(self.__fw.id())
                iface = self.__fw.get_iface_by_id(int(conn['iface']))
                if conn['addr']:
                    ndst_iface_addr = iface.get_addr_by_id(conn['addr'])
                else:
                    ndst_iface_addr = iface.addr()
                # net_src:
                conn = n['net_src'].get_conn(self.__fw.id())
                iface = self.__fw.get_iface_by_id(int(conn['iface']))
                if conn['addr']:
                    nsrc_iface_addr = iface.get_addr_by_id(conn['addr'])
                else:
                    nsrc_iface_addr = iface.addr()
                tunnelsrc = n['net_src'].remote()
                tunneldst = nsrc_iface_addr
                dir = 'in'
                #remote vpn network is destination of the connexion
                if n['net_dst'].is_vpn():
                    tunnelsrc = ndst_iface_addr
                    tunneldst = n['net_dst'].remote()
                    dir = 'out'
                t.extend(['--dir',dir, '--tunnel-src', tunnelsrc, '--tunnel-dst', tunneldst])
            if (n.get('net_src') is not None and self.__fw.is_default_net(n['net_src'])) or (n.get('net_dst') is not None and self.__fw.is_default_net(n['net_dst'])):
                has_internet = True
            #for INPUT/OUTPUT, don't put -i -o
            if not output: t.extend(['-i',n['if_src'].name().split(':')[0]])
            if not input: t.extend(['-o',n['if_dst'].name().split(':')[0]])
            t.extend(['-j',n['name']])
            st=gen_rule2(t)+os.linesep
            if n.get('net_src') is not None and n.get('net_dst') is not None and n['net_src'].is_vpn() and n['net_dst'].is_vpn():
                st=''
            if has_vpn: #we need vpn related flux to be first
                svpn+=st
            else:
                if has_internet and has_vpn is False:
                    if input:
                        sinout_internet += st
                        h_ininternet[n['net_src'].addr()] = st
                    elif output:
                        sinout_internet += st
                        l_outinternet.append(st)
                    else:
                        sinternet+=st
                        h_internet = self.insert_hash_net_chains(h_internet, n['net_src'].addr(), n['net_dst'].addr(), st)
                else: #put rules in order: forward, input, output
                    if input:
                        sinput += st
                        h_input[n['net_src'].addr()] = st
                    elif output:
                        soutput += st
                        l_output.append(st)
                    else:
                        s+=st
                        h_forward = self.insert_hash_net_chains(h_forward, n['net_src'].addr(), n['net_dst'].addr(), st)
        # rules for INVALID state
        invalid=self.__gen_default_invalid_drop(*lwhere)+os.linesep
        # LOG and DROP end rules
        default_log_drop = self.__gen_default_log_drop(*lwhere)
        sloopback = self.gen_loopback_default_rules(input = self.__fw.manage_input, output = self.__fw.manage_output)

        #get sorted rules for FORWARD, INPUT and OUTPUT
        s_sorted_forward = self.get_ipt_string(h_forward)
        input_keys = h_input.keys()
        input_keys.sort()
        s_sorted_input = ""
        for input_key in input_keys:
            s_sorted_input += h_input[input_key]
        s_sorted_output = ''.join(l_output)
        s_sorted_internet = self.get_ipt_string(h_internet)
        ininternet_keys = h_ininternet.keys()
        ininternet_keys.sort()
        s_sorted_ininternet = ""
        for ininternet_key in ininternet_keys:
            s_sorted_ininternet += h_ininternet[ininternet_key]
        s_sorted_outinternet = ''.join(l_outinternet)
        #res = estrel + invalid + svpn + s_sorted_forward + s_sorted_input +\
        #      s_sorted_output + s_sorted_ininternet + s_sorted_outinternet +\
        #      s_sorted_internet + os.linesep + sloopback + os.linesep +\
        #      default_log_drop + os.linesep
        #return res
        return estrel + invalid + svpn + s_sorted_forward + s_sorted_input +\
               s_sorted_output + s_sorted_ininternet + s_sorted_outinternet +\
               s_sorted_internet + os.linesep + sloopback + os.linesep +\
               default_log_drop + os.linesep

    def insert_hash_net_chains(self, hash, src, dst, mystring):
        """hash indexed by src network, dst network"""
        if not hash.has_key(src):
            hash[src] = {}
        hash[src][dst] = mystring
        return hash

    def get_ipt_string(self, hash):
        """get iptables commandes stored in hash
           commands are sorted by src_addr, dst_addr
        """
        ret = ''
        list_dst = hash.keys()
        list_dst.sort()
        for src_hash_key in list_dst:
            list_dst = hash[src_hash_key].keys()
            list_dst.sort()
            for dst_key in list_dst:
                cmd = hash[src_hash_key][dst_key]
                ret += cmd
        return ret


    def gen_rules_chain(self,chain,rescue=0):
        """Generate rules in one bichain"""
        s=''
        s2=''
        s3=''
        file_s = StringIO()
        file_s2 = StringIO()
        l=[]
        l2=[] #rules syn, rescue: put after new/rst/fin (l list)
        l3=[]
        l_eld = []
        rl = self.__fw.hrules()[chain]
        lch=rl.ordered_rules()
        if rl.is_bad():
            return FWipt.dumb_rules(lch,chain)
        #lch.sort()
        estrel=''
        sacl=''
        stmp=''
        comment=0
        has_rules = False
        queue_num = None
        # set queue_num if queue type is NFQUEUE
        if self.queue_type == D_NFQUEUE and self.queue_num:
            queue_num = self.queue_num
        for (rule_order, r) in lch:
            stmp=r.aclname; has_rules = True
            comment_rule=''
            if(stmp!=sacl):
                sacl=stmp
                sc=os.linesep+'#'+sacl+' (acl '+r.aclid+')'+os.linesep
                comment=1
            else: comment=0
            iface = ''
            if r.elt_dest().get('is_local'):
                iface = self.__fw.get_local_ip_connected_to(r.netsrc.addr())['iface']
            if r.from_nufw():# and r.proto().get('proto') in ['tcp']:#generate QUEUE cmds as in the Misc Article
                if r.proto().get('proto') not in ('tcp', 'udp'):#we comment bogus rules to avoid iptables errors and let client learn
                    comment_rule='#'
                    l.append('#ERROR_PROTO_AUTH'+os.linesep)
                if rescue==0:
                    if comment==1:
                        #s+=sc
                        #file_s.write(sc)
                        l.append(sc)
                    if r.is_auto == 1: l.append('#rules is auto'+os.linesep)
##                    inv=r.invert()
##                    invchain=self.__fw.chain(r.elt_dest().get('numnet'),r.elt_src().get('numnet'),r.elt_dest().get('zone'),r.elt_src().get('zone'),mark_src=r.elt_src().get('mark'),mark_dst=r.elt_dest().get('mark'))
##                    #generate session tracking rules if queue_num is set
##                    if not queue_num:
##                        sinv=str(RulePrintIptSimple(inv,chain=invchain['chain'],target=self.queue_type,insert='-A',add='-m state --state ESTABLISHED --tcp-flags SYN,ACK,RST,FIN SYN,ACK'))+os.linesep
##
##                    #s+=sinv
##                    #file_s.write(sinv)
##                        rst=RulePrintIptSimple(r,chain=chain,target=self.queue_type,add='-m state --state ESTABLISHED --tcp-flags RST RST')
##                        fin=RulePrintIptSimple(r,chain=chain,target=self.queue_type,add='-m state --state ESTABLISHED --tcp-flags FIN FIN')
##                        l.extend([comment_rule+sinv,comment_rule+str(rst)+os.linesep,comment_rule+str(fin)+os.linesep])
                    syn_str = "--syn"
                    if r.proto().get('proto') != 'tcp':
                        syn_str = ""
                    ript = RulePrintIptSimple(r,fw=self.__fw,chain=chain,target=self.queue_type,iface=iface, add='-m state --state NEW %s'%(syn_str), ulog_prefix=r.ulog_prefix)
                    if queue_num:
                        ript.set_queue_num(queue_num)
                    ssyn = str(ript)+os.linesep
                    if comment==1:
                        l2.append(sc)
                    l7mark = r.proto().get('l7mark')
                    l7mask = r.proto().get('l7mask')
                    if l7mark and not r.elt_dest().get('is_local'):
                        ript.target = D_CONNMARK
                        ript.mark_value = "%s/%s" %(l7mark, l7mask)
                        l2.append(str(ript)+os.linesep)
                    l2.append(comment_rule+ssyn)
                else:
                    rescue=RulePrintIptSimple(r,fw=self.__fw,chain=chain,iface=iface, add=_make_syn(make_target(r.decision),{'proto':r.proto().get('proto')}),target=make_target(r.decision),ulog_prefix=r.ulog_prefix)
                    srescue=str(rescue)+os.linesep
                    if comment==1:
                        #s2+=sc
                        #file_s2.write(sc)
                        l2.append(sc)
                    #s2+=srescue
                    #file_s2.write(srescue)
                    if r.is_auto == 1: l2.append('#rules is auto'+os.linesep)
                    l7mark = r.proto().get('l7mark')
                    l7mask = r.proto().get('l7mask')
                    if l7mark and not r.elt_dest().get('is_local'):
                        rescue.target = D_CONNMARK
                        rescue.mark_value = "%s/%s" %(l7mark, l7mask)
                        l2.append(str(rescue)+os.linesep)
                    l2.append(srescue)
            else:#generate normal iptables cmds: target depends on _decision_ attribut
                rstd=RulePrintIptSimple(r,fw=self.__fw,chain=chain,iface=iface, add=_make_syn(make_target(r.decision),{'proto':r.proto().get('proto')}),target=make_target(r.decision), ulog_prefix=r.ulog_prefix)
                std=str(rstd)
                if comment==1:
                    #s2+=sc
                    #file_s2.write(sc)
                    l2.append(sc)
                #s2+=std+os.linesep
                #file_s2.write(std+os.linesep)
                l7mark = r.proto().get('l7mark')
                l7mask = r.proto().get('l7mask')
                if l7mark and not r.elt_dest().get('is_local'):
                    rstd.target = D_CONNMARK
                    rstd.mark_value = "%s/%s" %(l7mark, l7mask)
                    l2.append(str(rstd)+os.linesep)
                l2.append(std+os.linesep)
        relaccept=''
        # Deleted on 2006-01-22 (ESTABLISHED are ACCEPT on a more global level)
        #if(rescue==0):#backup mode: don't need REL/EST command per chain
        #    relaccept=gen_rule('-A',chain,'-m state --state ESTABLISHED','-j ACCEPT')+os.linesep

        log=gen_rule('-A',chain,'-j',self.__logtype,self.__logprefix,'"%s%s"'%(chain,self.__logspace))+os.linesep
        drop=gen_rule('-A',chain,'-j DROP')

        if has_rules: l.append(2*os.linesep); l.extend(l2)
        l_eld.extend([relaccept,log,drop])
        return (l,l_eld)

    def gen_rules(self,rescue=0):
        """Generate rules for all bichains"""
        s=''; sep = os.linesep*2
        estrel=''
        l=[]; l_eld = []
        linput = []
        linput_eld = []
        loutput = []
        loutput_eld = [] # End Log Drop
        seen_chains = [] #used to not generate log/drop rules twice for a chain that exist with iface addr and broadcast
        for ch in self.__fw.chains(all=1,inout=1):# we make INPUT,OUTPUT and FORWARD rules in one step
            if ch['name'] in seen_chains:
                continue
            do_input = False; do_output = False
            (s,s_eld) = self.gen_rules_chain(ch['name'],rescue=rescue)
            #
            if(self.__fw.chain_input(ch)):
                if s != []:
                    linput.extend(s)
                    linput.append(sep)

                if ch['name'] not in seen_chains:
                    linput_eld.extend(s_eld)
                    linput_eld.append(sep)
                do_input = True
            if(self.__fw.chain_output(ch)):
                if s != []:
                    loutput.extend(s)
                    loutput.append(sep)
                if ch['name'] not in seen_chains:
                    loutput_eld.extend(s_eld); loutput_eld.append(sep)
                do_output = True
            if do_input is False and do_output is False:
                if s != []:
                    l.extend(s); l.append(sep)
                l_eld.extend(s_eld); l_eld.append(sep)
            #add chain to known chains
            if ch['name'] not in seen_chains:
                seen_chains.append(ch['name'])

        linput.extend(linput_eld)
        loutput.extend(loutput_eld)
        l.extend(l_eld)
        return (linput, loutput, l)


    def gen_vpn_rules(self):
        l = []
        for iface in self.__fw.ifaces():
            for n in iface.allnets():
                if n.is_vpn():
                    l.append('#%s%s'%(n.name(),os.linesep))
                    l.append(gen_rule('-A PREROUTING','-t mangle','-i',iface.name(),'-p 50','-s',n.remote(),'-j MARK --set-mark',n.mark())+os.linesep)
                    l.append(gen_rule('-A PREROUTING','-t mangle','-d',str(n.addr()),'-j MARK --set-mark',n.mark())+os.linesep)
        return l

    def gen_nat_rules(self):
        """Generate SNAT, DNAT and PNAT iptables rules
        """
        l = []
        #l.append(':nat%s'%os.linesep)
        snat,dnat,pnat = self.__fw.get_nat_rules()
        l.append('# SNAT%s'%os.linesep)
        def cmp_id(r1, r2):
            return cmp(r1.id, r2.id)
        snat.sort(cmp_id)
        dnat.sort(cmp_id)
        pnat.sort(cmp_id)
        for r in snat:
            l.append('# rule %s %s%s'%(str(r.id), r._name, os.linesep))
            l.append(str(SnatIpt(r))+os.linesep)
        l.append('# DNAT%s'%os.linesep)
        for r in dnat:
            l.append('# rule %s %s%s'%(str(r.id), r._name, os.linesep))
            l.append(str(DnatIpt(r))+os.linesep)
        l.append('# PNAT%s'%os.linesep)
        for r in pnat:
            l.append('# rule %s %s%s'%(str(r.id), r._name, os.linesep))
            l.append(str(PnatIpt(r))+os.linesep)
        #l.append('COMMIT%s'%os.linesep)
        return l

    def mangle_rules(self):
        """Create mangle rules for firewall"""
        # List of rules
        lrules = []
        """iptables -A PREROUTING -t mangle -j CONNMARK --restore-mark
> iptables -A POSTROUTING -t mangle -m mark ! --mark 0 -j CONNMARK
> --save-mark
>
>
        """
        if self.__fw.queue is not None:
            lrules.append(gen_rule('-A PREROUTING','-j CONNMARK --restore-mark',os.linesep))
            lrules.append(gen_rule('-A POSTROUTING', '-m mark', '! --mark 0/0x0000FFFF', '-j CONNMARK --save-mark --mask 0x0000FFFF', os.linesep))

        return lrules

#End class FWipt

class RulePrint:
    def __init__(self,r,fw='',net_from='',net_to=''):
        self.rule=r
        self.fw=fw
        self.ef=self.rule.elt_src()
        self.et=self.rule.elt_dest()
        self.do_vpn = True
        #if net_from!='':
            #self.net_from=net_from
        #else
            #self.net_from=self.ef.get('numnet')
           #
        #if net_to!='':
            #self.net_to=net_to
        #else
            #self.net_to=self.et.get('numnet')

    def __str__(self):
        pass

class RulePrintIptSimple(RulePrint):
    """produces a iptables command form an acl rule
    """
    #TODO: take chain and target in argument
    def __init__(self, r,fw='',chain='',target='DROP',net_from='',net_to='',add='',insert='', iface='', op_mark='set', mark='', ulog_prefix=''):
        #super(RulePrintIpt, self).__init__(r)
        RulePrint.__init__(self, r, fw,net_from=net_from,net_to=net_to)
        self.target=target
        self.chain=chain
        self.add=add
        self.insert='-A'
        self.queue_num = None
        if insert != '':
            self.insert=insert
        #if netsrc is not None: self.netsrc=netsrc
        #else:
        self.netsrc=self.rule.netsrc
        #if netdst is not None: self.netdst=netdst
        #else:
        self.netdst=self.rule.netdst
        self.iface=iface
        if ulog_prefix:
            self.ulog_prefix="%s:%s" %(r.aclid, ulog_prefix)
        else:
            self.ulog_prefix="%s:%s" %(r.aclid, self.chain)
        self.op_mark = op_mark
        self.mark_value = mark
        #self.dir=dir

    def _net_is_inclued_in(self,inner,net):
        """check if inner is inclued in net
        """
        if net is None: return False
        if inner.find('/')==-1: return False
        if net.zone()=='external': return False
        if net.zone()=='vpn': return False
        ntest=IP(inner)
        if ntest == net.addr(): return False
        conn = net.get_conn(self.fw.id())
        if ntest != IP('0.0.0.0/0') and "dftgateway" in conn:
            return True
        if ntest in net.addr():
            return True
        return False
    def set_ulog_prefix(self,s):
        self.ulog_prefix=s

    def set_queue_num(self, queue_num):
        self.queue_num = queue_num

    def __str__(self):
        #return "str RulePrintIpt"
        ef=self.rule.src().elts[0]
        et=self.rule.dest().elts[0]
        pr=self.rule.proto()
        t=[self.insert, self.chain]
        proto=''
        dport=''
        sport=''
        icmptype=''
        sip=''
        dip=''
        mark=''
        print_sip=False
        print_dip=True
        if pr.get('proto')!='':
            proto=pr.get('proto')
        if proto!='':
            t.extend(['-p',proto])

        if pr.get('icmptype')!='':
            icmptype=pr.get('icmptype')
        if pr.get('dport')!='':
            dport=pr.get('dport')
        if pr.get('sport')!='':
            sport=pr.get('sport')
        if et.get('mark')!='':
            mark=et.get('mark')
        if ef.get('mark')!='':
            mark=ef.get('mark')

        if icmptype!='':
            t.extend(['--icmp-type',icmptype])
        if dport!='':
            t.extend(['--dport',dport])
        if sport!='':
            t.extend(['--sport',sport])

        if self.iface:
            t.extend(['-i', self.iface.name()])

        #do we need to print the src/dst ip or dst/src net?
        if ef.get('numnet')!='':
            sip=ef.get('numnet')
        if et.get('numnet')!='':
            dip=et.get('numnet')
        if sip!='' and sip.find('/')==-1:
            #t.extend(['-s',sip])
            print_sip=True
        if dip!='' and dip.find('/')==-1:
            #t.extend(['-d',dip])
            print_dip=True
        if sip.find('/')!=-1:
            if self._net_is_inclued_in(sip,self.netsrc):
                print_sip=True
        if dip.find('/')!=-1:
            if self._net_is_inclued_in(dip,self.netdst):
                print_dip=True

        if mark!='' and self.do_vpn:
            t.extend(['-m mark','--mark',mark])
        if print_sip:
            t.extend(['-s',sip])
        if print_dip:
            t.extend(['-d',dip])

        t.extend([self.add,'-j',self.target])
        if self.target==D_ULOG and self.ulog_prefix!='':
            t.extend(['--ulog-prefix','"%s"'%(self.ulog_prefix)])
        if self.target==D_LOG and self.ulog_prefix!='':
            t.extend(['--log-prefix','"%s "'%(self.ulog_prefix)])
        if self.target==D_CONNMARK:
            t.extend(['--set-mark', '%s'%(self.mark_value)])
        if self.target==D_MARK:
            t.extend(['--%s-mark'%(self.op_mark), '%s'%(self.mark_value)])
        #set queue num if NFQUEUE is used
        if self.target == D_NFQUEUE and self.queue_num:
            t.append('--queue-num %s'%self.queue_num)

        t.extend([' #',str(self.rule.aclid)])
        t.extend([str(self.rule.cpt)])

        s2=gen_rule2(t)
        return s2



class GenNatIpt(object):
    """Produce iptables nat rules"""
    def __init__(self, rule):
        self._r = rule
        self._chain = ''
        self._target = ''
        self._targ = ''

    def _port_range(self, ports):
        if ports['l'] == ports['h'] and ports['h'] != -1:
            return [str(ports['h'])]
        elif ports['l'] != -1 and ports['h'] != -1:
            #return "%s:%s"%(ports['l'],ports['h'])
            return [str(ports['l']),str(ports['h'])]

    def __str__(self):
        #l = ["-t nat -A %s"%(self._chain)]
        l = ["-A %s"%(self._chain)]
        proto = self._r._proto
        if self._r._iface and self._ifacearg:
            l.append('%s %s'%(self._ifacearg, self._r._iface.split(':')[0]))
        if self._r.proto_defined():
            s_proto = ip_protocols[proto]
            l.append("-p %s"%s_proto)
        if self._r._src:
            l.append("-s %s"%(self._r._src))
        if self._r._dst:
            l.append("-d %s"%(self._r._dst))
        if proto in [PROTO_TCP, PROTO_UDP]:
            if self._r.port_defined(self._r._sport):
                l.append("--sport %s"%(':'.join(self._port_range(self._r._sport))))
            if self._r.port_defined(self._r._dport):
                l.append("--dport %s"%(':'.join(self._port_range(self._r._dport))))
        if proto == PROTO_ICMP:
            l.append("--icmp-type %s"%self._r.icmptype)
        l.append("-j %s"%self._target)
        to = self._r._to
        if self._r.port_defined(self._r._rport):
            to = "%s:%s"%(to,"-".join(self._port_range(self._r._rport)))
        if self._targ:
            l.append("%s %s"%(self._targ,to))
        return gen_rule2(l)

    #def cmp_id(self, r1, r2):
        #return cmp(r1.id, r2.id)
        #cmp_id = classmethod(cmp_id)



# classes for iptables NAT rules
class SnatIpt(GenNatIpt):
    def __init__(self, rule):
        super(SnatIpt, self).__init__(rule)
        self._chain = 'POSTROUTING'
        self._target = 'SNAT'
        self._targ = "--to-source"
        self._ifacearg = '-o'

class DnatIpt(GenNatIpt):
    def __init__(self, rule):
        super(DnatIpt, self).__init__(rule)
        self._chain = 'PREROUTING'
        self._target = 'DNAT'
        self._targ = "--to-destination"
        self._ifacearg = '-i'

class PnatIpt(GenNatIpt):
    def __init__(self, rule):
        super(PnatIpt, self).__init__(rule)
        self._chain = 'PREROUTING'
        self._target = 'DNAT'
        self._targ = '--to-destination'
        self._ifacearg = '-i'
        #self._target = None

if __name__ == "__main__":
    dnat = nunat.Dnat('192.168.2.2','10.10.2.3', 6,'10.3.6.6', sport='1024:65536', dport='80',id=1)
    snat = nunat.Snat('192.168.2.2','10.10.2.3', 6,'107.102.10.1', sport='1024:65536', dport='22',id=2)
    print SnatIpt(dnat)
    print DnatIpt(snat)


