#
# Copyright(C) 2006 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


"""XML parsing functions"""

import nuacl



def parse_periods(xmldoc):
    """ Get periods """
    periods = xmldoc.getElementsByTagName('periods')
    listpobj=nuacl.EltGrpList() #object list of periods
    if periods:
        period_node = periods[0]
        periods_nodes = period_node.getElementsByTagName('period')
        for period_n in periods_nodes:
            elts = period_n.getElementsByTagName('elt')
            try:
                mygrp = nuacl.xml_new_eltgrp(period_n.getAttribute('ID'),period_n.getAttribute('name'),period_n.getAttribute('op'),elts)
                listpobj.add(mygrp)
            except ValueError, e:
                print e
                sys.exit(1)
        return listpobj
    else:
        raise ValueError('no periods tag')


def parse_net_acls(xmldoc):
    """Parse nets tags in acls files"""
    res = {}
    descsn = xmldoc.getElementsByTagName('descs')
    if descsn is None:
        raise ValueError('no descs tags')
    descnodes = descsn[0].getElementsByTagName('desc')
    for descnode in descnodes:
        id = descnode.getAttribute('ID')
        if id == '':
            continue
        id = int(id)
        res[id] = {'id':id, 'name':descnode.getAttribute('name'), 'nets':{}}
        nets = {}
        for netnode in descnode.getElementsByTagName('net'):
            id_net = int(netnode.getAttribute('ID'))
            hnet = {'id':id_net, 'addr':netnode.getAttribute('addr'), 'name': netnode.getAttribute('name')}
            nets[id_net] = hnet
        res[id]['nets'] = nets
    return res


def parse_descsort(acl_node, descsort_list=None):
    """Parse descsort tags from acls xml file"""
    if descsort_list is None:
        descsort_list = []
    descsorts = acl_node.getElementsByTagName('descsorts')
    acl_id = acl_node.getAttribute('ID')
    acl_name = acl_node.getAttribute('name')
    #xml validity
    if descsorts is None:
        raise ValueError('no descsorts node for acl %s (%s)'%(acl_id, acl_name))
    descsort_nodes = descsorts[0].getElementsByTagName('descsort')
    if descsort_nodes is None:
        raise ValueError('no descsort node for acl %s (%s)'%(acl_id, acl_name))
    #get values
    res = {}
    for descsort_node in descsort_nodes:
        id_desc = descsort_node.getAttribute('ID')
        res[id_desc] = {}
        for elt_node in descsort_node.getElementsByTagName('elt'):
            elt_from = int(elt_node.getAttribute('from'))
            elt_to = int(elt_node.getAttribute('to'))
            elt_order = int(elt_node.getAttribute('order'))
            unit = (int(id_desc), elt_from, elt_to, elt_order)
            if unit in descsort_list:
                raise ValueError('Duplicated descsort node <ID=%s from=%s to=%s order=%s>'%unit)
            else:
                descsort_list.append(unit)
            cpl = (elt_from, elt_to)

            res[id_desc][cpl] = elt_order

    return res


if __name__ == "__main__":
    import sys
    import xml.dom.minidom

    #1er argument: fichier acls
    doc = xml.dom.minidom.parse(sys.argv[1])
    nets = parse_net_acls(doc)
    print nets
    for acl in doc.getElementsByTagName('acl'):
        descsorts = parse_descsort(acl)
        print "Acl %s: %s"%(acl.getAttribute('ID'), str(descsorts))


