<?php
/*
 * @copyright (C) 2004-2006 INL
 * @license http://www.gnu.org/licenses/gpl.txt GNU General Public License v.2
 * @author Damien Boucard <damien.boucard AT inl.fr>
 */

#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

require_once 'pear/Net/IPv4.php';

/*
 * Checks whether a network is included in another network
 * @author Vincent Deffontaines <gryzor@inl.fr>
 */
function net_in_network($net1,$mask1,$net2,$mask2) {
  $ip_calc = new Net_IPv4();
  $ip_calc->ip = $net1;
  if (preg_match("/\./",$mask1))
    $ip_calc->netmask = $mask1;
  else
    $ip_calc->bitmask = $mask1;
  $ip_calc->calculate();
  if (!Net_IPv4::ipInNetwork($ip_calc->network, $net2."/".$mask2))
  {
    return(0);
  }
  if (!Net_IPv4::ipInNetwork($ip_calc->broadcast, $net2."/".$mask2))
  {
    return(0);
  }
  return(1);
}

/*
 * Parses a string representing an network address (IP with netmask or IP with
 * bitlength) to an array containing separately the IP address and the netask.
 *
 * @author Vincent Deffontaines <gryzor@inl.fr>
 */
function string_to_net($s) {
  $splitted = preg_split("/\//",$s);
  if (count($splitted)>1)
    return ($splitted);
  if ($splitted[0] == "0.0.0.0")
  {
    array_push($splitted,"0.0.0.0");
    return($splitted);
  }
  array_push($splitted,"255.255.255.255");
  return($splitted);
}
    
/*
 * This class sort by insertion a list of netmask addresses.
 *
 * Note: This sort only guarantees that an included network
 * is before its including network.
 */
class NetmaskList {
  var $head;

  function NetmaskList($array=null) {
    $this->void = null;
    $this->head = $this->void;
    if ($array) {
      foreach ($array as $addr=>$elt) {
        $this->addNetmask($addr, $elt);
      }
    }
  }

  function addNetmask($addr, $data) {
    if ($this->head == $this->void) {
      $this->head = &new NetmaskElement($addr, $data, $this->void, $this->void);
      return;
    }
    $iterator = &$this->head;
    while (1) {
      # Check if $addr is a network included in the current iterated network
      $inc_net_tab = string_to_net($addr);
      $over_net_tab = string_to_net($iterator->address);
      if (net_in_network($inc_net_tab[0], $inc_net_tab[1], $over_net_tab[0], $over_net_tab[1])) {
        $element = &new NetmaskElement($addr, $data, $iterator->previous, $iterator);
        $iterator->previous->next = &$element;
        $iterator->previous = &$element;
        if ($iterator == $this->head) {
          $this->head = &$iterator->previous;
        }
        return;
      }
      # End of the list : we insert not-included network
      if ($iterator->next == $this->void) {
        $iterator->next = &new NetmaskElement($addr, $data, $iterator, $this->void);
        return;
      }
      # Iteration
      $iterator = &$iterator->next;
    }
  }

  function getArray() {
    $result = array();
    for ($iterator=$this->head; $iterator!=$this->void; $iterator=$iterator->next) {
      $result[$iterator->address] = $iterator->data;
    }
    return $result;
  }
}

class NetmaskElement {
  var $address;
  var $data;
  var $previous;
  var $next;

  function NetmaskElement($address, $data, &$previous, &$next) {
    $this->address = $address;
    $this->data = $data;
    $this->previous = &$previous;
    $this->next = &$next;
  }
}

function netmask_test(){
  $data_set = array("192.168.33.0/25"=>1, "192.168.33.128/25"=>2, "10.0.0.0/8"=>3, "192.168.0.0/16"=>4, "10.51.0.0/16"=>5, "10.52.0.0/16"=>6, "0.0.0.0/0"=>7, "195.192.53.1/32"=>8, "192.168.33.1/32"=>9);
  print 'Base of data:';
  print '<pre>'; print_r($data_set); print '</pre>';
  $list = new NetmaskList();
  print 'Step-by-step insertion:';
  print '<pre>'; print_r($list->getArray()); print '</pre>';
  foreach ($data_set as $addr=>$data) {
    $list->addNetmask($addr, $data);
    print '<pre>'; print_r($list->getArray()); print '</pre>';
  }
  print 'Array insertion:';
  $list2 = new NetmaskList($data_set);
  print '<pre>'; print_r($list2->getArray()); print '</pre>';
}