#!/usr/bin/env python
import sys, re, os

ignore_regex = re.compile(r'^dn: ou=acls,dc=test,dc=lld')

regex_string = r'''^dn: cn=(?P<cn>[a-zA-Z0-9_ \-\.]+_[0-9]+_(?P<random5>[0-9]+)_(?P<random6>[0-9]+)),ou=acls,dc=test,dc=lld
DstPortStart: \d+
cn: (?P=cn)
Proto: (6|17)
objectClass: top
objectClass: NuAccessControlList
Decision: (0|1)
DstPortEnd: \d+
Group: [0-9]+
description: \d+:
(SrcPortEnd: \d+\n)?DstIPEnd: \d+
SrcIPEnd: \d+
DstIPStart: \d+
SrcIPStart: \d+
(SrcPortStart: \d+\n)?DstPort: \d+
structuralObjectClass: NuAccessControlList
entryUUID: (?P<random1>[0-9a-f\-]+)
creatorsName: cn=admin,dc=test,dc=lld
createTimestamp: (?P<random3>\d{14}[A-Z])
entryCSN: (?P<random2>\d{14}[A-Z][#0-9a-f]{17})
modifiersName: cn=admin,dc=test,dc=lld
modifyTimestamp: (?P<random4>\d{14}[A-Z])
$'''

NB_RANDOM_FIELDS = 6
DEBUG_MODE = False

entry_regex = re.compile(regex_string)

def compare_block(entry, ref):
    if ignore_regex.match(entry):
        return None
    match = entry_regex.match(entry)
    if match is None:
        if DEBUG_MODE:
            import StringIO
            regex_file = StringIO.StringIO(regex_string)
            entry_file = StringIO.StringIO(entry)
            while True:
                try:
                    regex = regex_file.next()
                    entry = entry_file.next()
                    regex = regex.replace("(?P=cn)", ".*")
                    m = re.match(regex, entry)
                    if m is None:
                        prefix = "**"
                    else:
                        prefix = "OK"
                    print "%s %r | %r" %(prefix, regex, entry)
                except StopIteration:
                    break
        
        return "Entry does not match pattern"
    group_dict = match.groupdict()
    for i in range(1,NB_RANDOM_FIELDS+1):
        val = group_dict["random%s"%i]
        entry = entry.replace(val, '#####')
    if entry == ref:
        return None
    else:
        return entry

def test_ldap():
    if not os.access(sys.argv[1], os.F_OK):
        sys.exit(0)
    ref_file = open(sys.argv[1], 'r')
    ldap_contents = sys.stdin.read()
    ldap_contents = ldap_contents.replace("\n ", "")
    ldap_lines = ldap_contents.split("\n")
    ldap_entry = []
    ref_entry = []
    for ldap_line in ldap_lines:
        ldap_line += "\n"
        try:
            ref_line = ref_file.next()
        except StopIteration:
            ref_line = '\n'
	if DEBUG_MODE:
            print ">>> %r" %(ldap_line)
            print "### %r" %(ref_line)
        if ldap_line == "\n" and ref_line == "\n":
            if len(ldap_entry) == 0 and len(ref_entry) == 0:
                break
            entry = compare_block(''.join(ldap_entry), ''.join(ref_entry))
            if entry is not None:
                print >>sys.stderr, """
Entry does not match:
---------------------
%s
Expected:
---------
%s""" %(entry, ''.join(ref_entry))
                sys.exit(1)
            ldap_entry = []
            ref_entry = []
        elif ldap_line and ldap_line != "\n" and ref_line and ref_line != "\n":
            ldap_entry.append(ldap_line)
            ref_entry.append(ref_line)
        else:
            if len(ldap_entry) == 0:
                print >>sys.stderr, "No ACLs in LDAP database"
            else:
                print >>sys.stderr, "Entry size different for %s" %(ldap_entry[0])
            sys.exit(1)

def import_ldap():
    dest_file = open(sys.argv[1], 'w')
    dest_lines = []
    ldap_contents = sys.stdin.read()
    ldap_contents = ldap_contents.replace("\n ", "")
    ldap_lines = ldap_contents.split("\n")
    ldap_entry = []
    for ldap_line in ldap_lines:
        ldap_line += "\n"
        if ldap_line == "\n":
            if len(ldap_entry) == 0:
                break
            entry = ''.join(ldap_entry)
            ldap_entry = []
            if ignore_regex.match(entry):
                dest_lines.append(entry)
                continue
            match = entry_regex.match(entry)
            if match is None:
                print "Entry does not match pattern"
                if DEBUG_MODE:
                  print entry
                sys.exit(1)
            group_dict = match.groupdict()
            for i in range(1,NB_RANDOM_FIELDS+1):
                val = group_dict["random%s"%i]
                entry = entry.replace(val, '#####')
            dest_lines.append(entry)
        else:
            ldap_entry.append(ldap_line)
    dest_file.write("\n".join(dest_lines))
    dest_file.close()

if __name__ == "__main__":
    if sys.argv[-1] == 'import':
      import_ldap()
    else:
      test_ldap()

